#include <cstdio>
#include <cstring>

#include "gp32.h"

#include "gp_gfx.h"
#include "gp_io.h"
#include "gfx/chess_gfx.h"

extern "C"
{
	void EnableCache();
	void gp_setCpuspeed(int);
	void gp_disableIRQ();
	void gp_installSWIIRQ(int,void (*)(void));
	void gp_enableIRQ();
	void gp_drawString(int,int,int,char*,u16,void*);
	void gp_initButton();
	u16  gp_getButton();
	void gp_Reset();
}

typedef enum
{
	ROOK = 1,
	KNIGHT,
	BISHOP,
	QUEEN,
	KING,
	PAWN
} PIECE;

#define BLACK 0x80
#define PIECE(a) ((a)&0x7f)
#define COLOR(a) ((a)&BLACK)

const static int knight_mx[8] = {-1,-1,-2,-2, 2, 2, 1, 1};
const static int knight_my[8] = {-2, 2,-1, 1, 1,-1, 2,-2};

u8 board[64];
u8 bctrl[64];
u8 wctrl[64];
u8 bking,wking;
u8 bkmove1,wkmove1,bkmove2,wkmove2;
int passant;

#define DX 56
#define DY 16
#define SQUARE 26
#define ABS(a) ((a)<0 ? -(a) : (a))

int stepProto;
static char game[1024];
int curX,curY,oldX,oldY;
char* last;
int lastX0,lastY0,lastX1,lastY1;
int bflip;
int values[6] = {0,0,4,5,0,0};

int hcolor;
int mode;	// 0 : fixed depth, 1 : time
int fdepth;	// depth
int tincr;
int wtime,btime;
u32 time_t0;
u32 time_t1;
int cpu;

#define DELTA(a) (((a-time_t0)*1000)>>6)

// RTC implementation based on MIRKO's SDK code

volatile u32 myGLOBALCOUNTER;
volatile u8 myGLOBALFLAG;

static void myRTCInt(void) __attribute__ ((interrupt ("IRQ")));

static void myRTCInt(void)
{
	myGLOBALCOUNTER++;
	if ((myGLOBALCOUNTER&0x1f)==0)
		myGLOBALFLAG = 1;
}

static void mySetupRTCInt( void )
{
	rCLKCON |= 0x800;
	rTICINT = 0x81;
}

static void myInstallRTC( void )
{
	gp_disableIRQ();
	gp_installSWIIRQ(8,myRTCInt);
	gp_enableIRQ();
}

u32 gp_getMyRTC()
{
	return myGLOBALCOUNTER;
}

void gp_initMyRTC()
{
	myGLOBALCOUNTER = 0;
	myGLOBALFLAG = 0;
	mySetupRTCInt();
	myInstallRTC();
}

void gp_clearMyRTC()
{
	myGLOBALCOUNTER=0;
}

// End RTC code

void initBoard()
{
	int i;

	memset(board,0,64);
	memset(bctrl,0,64);
	memset(wctrl,0,64);
	board[0] = board[7] = board[56] = board[63] = ROOK;
	board[1] = board[6] = board[57] = board[62] = KNIGHT;
	board[2] = board[5] = board[58] = board[61] = BISHOP;
	board[3] = board[59] = QUEEN;
	board[4] = board[60] = KING;
	for (i=0;i<8;i++)
		board[8+i] = board[48+i] = PAWN;
	for (i=0;i<16;i++)
		board[48+i] |= BLACK;
	passant = -1;
	bking = 60;
	wking = 4;
	wkmove1 = wkmove2 = 0;
	bkmove1 = bkmove2 = 0;

	last = NULL;
	strcpy(game,"position startpos moves ");
	curX = oldX = -1;
	lastX0 = -1;
}

void copyBlock(u16* dst,int dx,int dy,int sx,int sy,int sw,int sh,const u16* src,int srcw,int srch)
{
	src += (sx*srch)+srch-sy-sh;
	dst += dx*240+dy;
	sh*=2;
	for (int i=0;i<sw;i++)
	{
		memcpy(dst,src,sh);
		src+=srch;
		dst+=240;
	}
}

void drawRect(int x,int y,int w,int h,u16 col)
{
	int i,j;
	u16* buf = getCurrentBuffer();
	buf+=x*240+240-y-h;
	w-=1;
	i = w*240;
	for (j=0;j<h;j++)
	{
		buf[j] = col;
		buf[i++] = col;
	}
	h-=1;
	buf+=240;
	for (i=1;i<w;i++)
	{
		buf[0] = col;
		buf[h] = col;
		buf+=240;
	}
}

void gp_mainMenu()
{
	const static char* greetz[] =
	{
		"Game Partner 32 v2.0",
		"Engine Fruit 2.1 by Fabien Letouzey",
		"http://arctrix.com/nas/fruit/",
		"http://www.fruitchess.com",
		"",
		"GP32 Port by [SOD]Thor",
		"",
		"Based on DSChess (Boris von Loesch)",
		"http://www.vonloesch.de/dschess",
		NULL
	};

	const static char* options[8] = {"Color","Mode","Depth","Minutes","Increment(s)","CPU","Play !","Exit"};

	const static char* vopts[2][2] = {{"WHITE","BLACK"},{"DEPTH","TIME"}};

	const static int vmin[6] = {0,0,1,2,0,0};
	const static int vmax[6] = {1,1,10,30,30,4};
	const static int vcpu[5] = {66,100,133,144,156};
	u16* buf;
	int key,y=6,i,j,dy;
	char strbuf[32];

	gp_setCpuspeed(40);

	console_write("Why do you press a button ?",COLOR_WHITE);
	while ((key=gp_getButton())!=0);

	clearScreen();
	while(true)
	{
		buf = getCurrentBuffer();
		memset(buf,0,320*240*2);
		gp_drawString(80,12,strlen((char*)greetz[0]),(char*)greetz[0],COLOR_WHITE,buf);
		for (i=0;i<8;i++)
		{
			dy = 36+i*14;
			gp_drawString(80,dy,strlen((char*)options[i]),(char*)options[i],COLOR_WHITE,buf);
			if (i<2)
				gp_drawString(200,dy,strlen(vopts[i][values[i]]),(char*)vopts[i][values[i]],COLOR_WHITE,buf);
			else if (i<6)
			{
				sprintf(strbuf,"%d",i!=5?values[i]:vcpu[values[i]]);
				gp_drawString(200,dy,strlen(strbuf),strbuf,COLOR_WHITE,buf);
			}
		}
		for (i=1;greetz[i]!=NULL;++i)
		{
			j = strlen((char*)greetz[i]);
			gp_drawString(160-4*j,144+i*10,j,(char*)greetz[i],COLOR_WHITE,buf);
		}
		drawRect(76,33+y*14,166,14,COLOR_WHITE);
		drawRect(10,8,300,16,COLOR_WHITE);
		drawRect(10,26,300,122,COLOR_WHITE);
		drawRect(10,150,300,86,COLOR_WHITE);
		flip(1);
		key=gp_getButton();
		if (key & (BUTTON_UP | BUTTON_LEFT | BUTTON_L))
			y = y==0 ? 7 : y-1;
		if (key & (BUTTON_DOWN | BUTTON_RIGHT | BUTTON_R))
			y = y==7 ? 0 : y+1;
		if (y==7 && (key & (BUTTON_A | BUTTON_B)))
			gp_Reset();
		if ((key & BUTTON_START) || (y==6 && (key & (BUTTON_A | BUTTON_B))))
			break;
		if (key & (BUTTON_A | BUTTON_SELECT))
			values[y] = (values[y]<vmax[y] ? values[y]+1 : vmin[y]);
		if (key & BUTTON_B)
			values[y] = (values[y]>vmin[y] ? values[y]-1 : vmax[y]);
		while ((key=gp_getButton())!=0);
	}
	mode = values[1];
	hcolor = values[0]==0 ? 0 : BLACK;
	bflip = hcolor==BLACK;
	fdepth = values[2];
	wtime = btime = values[3]*60000;
	tincr = values[4]*1000;
	cpu = vcpu[values[5]];
}

bool gp_subMenu()
{
	const static char* entries[] =
	{
		"Game Partner 32 v2.0",
		"Continue",
		"Abort game",
		"Exit",
		NULL
	};

	u16* buf;
	int key,y=0,i,j;

	clearScreen();
	while(true)
	{
		buf = getCurrentBuffer();
		memset(buf,0,320*240*2);
		for (i=0;entries[i]!=NULL;++i)
		{
			j = strlen((char*)entries[i]);
			gp_drawString(160-4*j,(i==0?40:100)+i*14,j,(char*)entries[i],COLOR_WHITE,buf);
		}
		drawRect(76,111+y*14,166,14,COLOR_WHITE);
		drawRect(10,36,300,16,COLOR_WHITE);
		drawRect(10,100,300,60,COLOR_WHITE);
		flip(1);
		key=gp_getButton();
		if (key & (BUTTON_UP | BUTTON_LEFT | BUTTON_L))
			y = y==0 ? 2 : y-1;
		if (key & (BUTTON_DOWN | BUTTON_RIGHT | BUTTON_R))
			y = y==2 ? 0 : y+1;
		if (key & (BUTTON_A | BUTTON_B | BUTTON_SELECT))
			break;
		while ((key=gp_getButton())!=0);
	}
	while ((key=gp_getButton())!=0);
	if (y==1)
	{
		stepProto = 6;
		return true;
	}
	else if (y==2)
		gp_Reset();
	return false;
}

void showPos(int x,int y,u16 col)
{
	x = DX+(bflip?7-x:x)*SQUARE;
	y = 32-DY+(bflip?y:7-y)*SQUARE;
	drawRect(x,y,SQUARE,SQUARE,col);
	drawRect(x+1,y+1,SQUARE-2,SQUARE-2,col);
}

void showBoard()
{
	int i,j,c,x,y;
	u16* buf = getCurrentBuffer();
	int w,b,wm,ws,bm,bs;
	u32 ct;
	char str[64];

	memset(buf,0,320*240*2);
	for (i=0;i<8;i++)
	{
		for (j=0;j<8;j++)
		{
			y = (i+j)&1;
			if (!(c = board[i+j*8]))
				x = 6;
			else
			{
				x = PIECE(c)-1;
				y += (COLOR(c)==BLACK ? 2 : 0);
			}
			copyBlock(buf,DX+(bflip?7-i:i)*SQUARE,DY+(bflip?7-j:j)*SQUARE,x*SQUARE,y*SQUARE,SQUARE,SQUARE,GFX_IMG,GFX_W,GFX_H);
		}
	}
	if (oldX<0 && lastX0>=0)
	{
		showPos(lastX0,lastY0,COLOR_MAGENTA);
		showPos(lastX1,lastY1,COLOR_MAGENTA);
	}
	if (curX>=0)
		showPos(curX,curY,COLOR_YELLOW);
	if (oldX>=0)
		showPos(oldX,oldY,COLOR_RED);
	if (stepProto==6)
		gp_drawString(DX,228,13,"ABORTING GAME",COLOR_WHITE,buf);
	else if (last!=NULL)
		gp_drawString(DX,228,strlen((char*)last),(char*)last,COLOR_WHITE,buf);
	if (bctrl[wking]||wctrl[bking])
		gp_drawString(6,116,5,"check",COLOR_WHITE,buf);
	if (mode==1 && (stepProto==4 || stepProto==5))
	{
		ct = gp_getMyRTC();
		if ((hcolor==BLACK && stepProto==4)||(hcolor!=BLACK && stepProto==5))
		{
			w = wtime;
			b = btime-DELTA(ct);
		}
		else
		{
			w = wtime-DELTA(ct);
			b = btime;
		}
		if (w<0)
			w = 0;
		if (b<0)
			b = 0;
		wm = w/60000;
		ws = (w-wm*60000)/1000;
		bm = b/60000;
		bs = (b-bm*60000)/1000;
		sprintf(str,"white %02d:%02d    black %02d:%02d",wm,ws,bm,bs);
		gp_drawString(DX,5,strlen(str),str,COLOR_WHITE,buf);
	}
	flip(1);
}

void timeUpdate(int* t)
{
	int delta = DELTA(time_t1);
	if (delta>*t)
		*t = 0;
	else
		*t = *t-delta+tincr;
}

bool gp_show()
{
	int key;
	bool ret = false;

	if (stepProto==6)
		return true;

	key=gp_getButton();

	if (mode==1 && myGLOBALFLAG==1)
	{
		if (key & (BUTTON_L | BUTTON_R))
			bflip ^= 1;
		showBoard();
		myGLOBALFLAG = 0;
	}

	if (key & BUTTON_START)
	{
		if (mode==1)
		{
			time_t1 = gp_getMyRTC();
			if (hcolor==BLACK)
				timeUpdate(&wtime);
			else
				timeUpdate(&btime);
		}
		ret = gp_subMenu();
		if (mode==1)
			time_t0 = gp_getMyRTC();
	}

	return ret;
}

void line_control(u8* ctrl,int x,int y,int dx,int dy)
{
	int k;
	x+=dx;
	y+=dy;
	while (x>=0 && x<8 && y>=0 && y<8)
	{
		k = x+(y<<3);
		ctrl[k] = 1;
		if (board[k]!=0)
			break;
		x+=dx;
		y+=dy;
	}
}

void playMove(int x0,int y0,int x1,int y1,int promo)
{
	int src = x0+(y0<<3);
	int dst = x1+(y1<<3);
	int p = PIECE(board[src]);
	int c = COLOR(board[src]);
	int i,j,x,y,kx,ky;
	u8* ctrl;

	// promotion + en passant
	if (p==PAWN)
	{
		switch(promo)
		{
			case 'q': p = QUEEN;  break;
			case 'r': p = ROOK;   break;
			case 'n': p = KNIGHT; break;
			case 'b': p = BISHOP; break;
		}

		if (x0!=x1 && board[dst]==0)
			board[x1+(y0<<3)] = 0;
		else if (y1-y0==2 || y1-y0==-2)
			passant = x0;
		else
			passant = -1;
	}
	else
		passant = -1;

	// do move
	board[dst] = p|c;
	board[src] = 0;

	// rook moved : no castle
	if (p==ROOK)
	{
		if (x0==0)
		{
			if (c==BLACK)
				bkmove1 = 1;
			else
				wkmove1 = 1;
		}
		else if (x0==7)
		{
			if (c==BLACK)
				bkmove2 = 1;
			else
				wkmove2 = 1;
		}
	}

	// castle
	if (p==KING)
	{
		if (src==4)
		{
			if (dst==6)
			{
				board[5] = board[7];
				board[7] = 0;
			}
			else if (dst==2)
			{
				board[3] = board[0];
				board[0] = 0;
			}
		}
		else if (src==60)
		{
			if (dst==62)
			{
				board[61] = board[63];
				board[63] = 0;
			}
			else if (dst==58)
			{
				board[59] = board[56];
				board[56] = 0;
			}
		}
		if (c==BLACK)
		{
			bkmove1 = bkmove2 = 1;
			bking = dst;
		}
		else
		{
			wkmove1 = wkmove2 = 1;
			wking = dst;
		}
	}

	// control

	memset(bctrl,0,64);
	memset(wctrl,0,64);

	for (i=0;i<64;i++)
	{
		p = PIECE(board[i]);

		if (p==0)
			continue;

		c = COLOR(board[i]);
		ctrl = (c==BLACK ? bctrl : wctrl);
		x = i&0x7;
		y = (i>>3);

		switch(p)
		{
			case ROOK:
				line_control(ctrl,x,y,-1, 0);
				line_control(ctrl,x,y, 1, 0);
				line_control(ctrl,x,y, 0,-1);
				line_control(ctrl,x,y, 0, 1);
				break;

			case KNIGHT:
				for (j=0;j<8;j++)
				{
					kx = x+knight_mx[j];
					ky = y+knight_my[j];
					if (kx<0 || kx>7 || ky<0 || ky>7)
						continue;
					ctrl[kx+(ky<<3)] = 1;
				}
				break;

			case BISHOP:
				line_control(ctrl,x,y, 1, 1);
				line_control(ctrl,x,y,-1, 1);
				line_control(ctrl,x,y, 1,-1);
				line_control(ctrl,x,y,-1,-1);
				break;

			case QUEEN:
				line_control(ctrl,x,y,-1, 0);
				line_control(ctrl,x,y, 1, 0);
				line_control(ctrl,x,y, 0,-1);
				line_control(ctrl,x,y, 0, 1);
				line_control(ctrl,x,y, 1, 1);
				line_control(ctrl,x,y,-1, 1);
				line_control(ctrl,x,y, 1,-1);
				line_control(ctrl,x,y,-1,-1);
				break;

			case KING:
				if (x>0)
				{
					ctrl[i-1] = 1;
					if (y>0)
						ctrl[i-9] = 1;
					if (y<7)
						ctrl[i+7] = 1;
				}
				if (y>0)
					ctrl[i-8] = 1;
				if (y<7)
					ctrl[i+8] = 1;
				if (x<7)
				{
					ctrl[i+1] = 1;
					if (y>0)
						ctrl[i-7] = 1;
					if (y<7)
						ctrl[i+9] = 1;
				}
				break;

			case PAWN:
				if (c==BLACK)
				{
					if (x>0)
						ctrl[i-9] = 1;
					if (x<7)
						ctrl[i-7] = 1;
				}
				else
				{
					if (x>0)
						ctrl[i+7] = 1;
					if (x<7)
						ctrl[i+9] = 1;
				}
		}
	}
}

bool noPath(int x0,int y0,int x1,int y1)
{
	int dx = x0==x1 ? 0 : (x0<x1?1:-1);
	int dy = y0==y1 ? 0 : (y0<y1?1:-1);
	
	x0+=dx;
	y0+=dy;

	while (x0!=x1 || y0!=y1)
	{
		if (board[x0+(y0<<3)]!=0)
			return true;
		x0+=dx;
		y0+=dy;
	}
	return false;
}

bool isMoveValid(int x0,int y0,int x1,int y1,int promo)
{
	int src = x0+(y0<<3);
	int dst = x1+(y1<<3);
	int p = PIECE(board[src]);
	int c = COLOR(board[src]);
	int p2 = board[dst];
	int c2 = COLOR(p2);
	int i;
	u8 save_board[64];
	u8 save_bctrl[64];
	u8 save_wctrl[64];
	u8 save_bking;
	u8 save_wking;
	u8 save_wkmove1;
	u8 save_wkmove2;
	u8 save_bkmove1;
	u8 save_bkmove2;
	u8 save_passant;

	// Color check
	if (p2!=0 && c==c2)
		return false;

	// Pawn move
	switch(p)
	{
		case ROOK:
			if ((x0!=x1 && y0!=y1) || noPath(x0,y0,x1,y1))
				return false;
			break;

		case KNIGHT:
			for (i=0;i<8;i++)
			{
				if (x1==x0+knight_mx[i] && y1==y0+knight_my[i])
					break;
			}
			if (i==8)
				return false;
			break;

		case BISHOP:
			if ((ABS(x1-x0)!=ABS(y1-y0)) || noPath(x0,y0,x1,y1))
				return false;
			break;

		case QUEEN:
			if (((x0==x1 && y0!=y1) || (x0!=x1 && y0==y1) || (ABS(x1-x0)==ABS(y1-y0))) && noPath(x0,y0,x1,y1))
				return false;
			break;

		case KING:
			if (c==BLACK && src==60 && dst==62 && bkmove2==0)
			{
				if (board[61]!=0 || board[62]!=0 ||
					wctrl[60]==1 || wctrl[61]==1 || wctrl[62]==1)
					return false;
			}
			else if (c==BLACK && src==60 && dst==58 && bkmove1==0)
			{
				if (board[59]!=0 || board[58]!=0 ||
					wctrl[60]==1 || wctrl[59]==1 || wctrl[58]==1)
					return false;
			}
			else if (c!=BLACK && src==4 && dst==6 && wkmove2==0)
			{
				if (board[5]!=0 || board[6]!=0 ||
					bctrl[4]==1 || bctrl[5]==1 || bctrl[6]==1)
					return false;
			}
			else if (c!=BLACK && src==4 && dst==2 && wkmove1==0)
			{
				if (board[3]!=0 || board[2]!=0 || 
					bctrl[4]==1 || bctrl[3]==1 || bctrl[2]==1)
					return false;
			}
			else if (ABS(x1-x0)>1 || ABS(y1-y0)>1)
				return false;
			break;

		case PAWN:
			if (c==BLACK)
			{
				if (x0==x1)
				{
					if (board[src-8]!=0)
						return false;
					if (y0==6 && y1==4)
					{
						if (p2!=0)
							return false;
					}
					else if (y1!=y0-1)
						return false;
				}
				else
				{
					if (y0!=y1+1)
						return false;
					if (p2==0) // en passant
					{
						if (y1!=2 || passant!=x1)
							return false;
					}
				}
			}
			else
			{
				if (x0==x1)
				{
					if (board[src+8]!=0)
						return false;
					if (y0==1 && y1==3)
					{
						if (p2!=0)
							return false;
					}
					else if (y1!=y0+1)
						return false;
				}
				else
				{
					if (y0!=y1-1)
						return false;
					if (p2==0) // en passant
					{
						if (y1!=5 || passant!=x1)
							return false;
					}
				}
			}
	}

	// test check	
	memcpy(save_board,board,64);
	memcpy(save_bctrl,bctrl,64);
	memcpy(save_wctrl,wctrl,64);
	save_bking = bking;
	save_wking = wking;
	save_wkmove1 = wkmove1;
	save_wkmove2 = wkmove2;
	save_bkmove1 = bkmove1;
	save_bkmove2 = bkmove2;
	save_passant = passant;

	playMove(x0,y0,x1,y1,promo);

	if ((c==BLACK && wctrl[bking]==0) || (c!=BLACK && bctrl[wking]==0))
		return true;

	memcpy(board,save_board,64);
	memcpy(bctrl,save_bctrl,64);
	memcpy(wctrl,save_wctrl,64);
	bking = save_bking;
	wking = save_wking;
	wkmove1 = save_wkmove1;
	wkmove2 = save_wkmove2;
	bkmove1 = save_bkmove1;
	bkmove2 = save_bkmove2;
	passant = save_passant;

	return false;
}

void computerMove(char *ucicmd)
{
	char buf[256];
	if (mode==1)
		sprintf(buf,"go btime %d wtime %d binc %d winc %d",btime,wtime,tincr,tincr);
	else
		sprintf(buf,"go depth %d",fdepth);
	gp_write(ucicmd);
	gp_write(buf);
	stepProto = 5;
	if (mode==1)
		time_t0 = gp_getMyRTC();
	showBoard();
	gp_setCpuspeed(cpu);
}

void humanMove()
{
	int key, p, pp;

	gp_setCpuspeed(40);

	if (mode==1)
		time_t0 = gp_getMyRTC();

	curX = lastX0 < 0 ? 4 : lastX1;
	curY = lastX0 < 0 ? (hcolor==BLACK ? 6 : 1) : lastY1;

	showBoard();
	while ((key=gp_getButton())!=0);

	while(true)
	{
		oldX = -1;
		p = 0;
		// choose piece to move
		while(p==0)
		{
			showBoard();
			key=gp_getButton();
			if (key & BUTTON_UP)
				curY = (curY+(bflip?-1:1))&7;
			if (key & BUTTON_DOWN)
				curY = (curY+(bflip?1:-1))&7;
			if (key & BUTTON_LEFT)
				curX = (curX+(bflip?1:-1))&7;
			if (key & BUTTON_RIGHT)
				curX = (curX+(bflip?-1:1))&7;
			if (key & (BUTTON_L | BUTTON_R))
				bflip ^= 1;
			if (key & BUTTON_A)
				p = board[curX+(curY<<3)];
			if (key & BUTTON_START)
			{
				if (gp_subMenu())
					return;
			}
			while ((key=gp_getButton())!=0);
		}
		if (COLOR(p)!=hcolor)
			continue;
		pp = PIECE(p);
		// choose destination
		oldX = curX;
		oldY = curY;
		while(p)
		{
			showBoard();
			key = gp_getButton();
			if (key & BUTTON_UP)
				curY = (curY+(bflip?-1:1))&7;
			if (key & BUTTON_DOWN)
				curY = (curY+(bflip?1:-1))&7;
			if (key & BUTTON_LEFT)
				curX = (curX+(bflip?1:-1))&7;
			if (key & BUTTON_RIGHT)
				curX = (curX+(bflip?-1:1))&7;
			if (key & (BUTTON_L | BUTTON_R))
				bflip ^= 1;
			if (key & BUTTON_A)
				p = 0;
			if (key & BUTTON_B)
				break;
			if (key & BUTTON_START)
			{
				if (gp_subMenu())
					return;
			}
			while ((key=gp_getButton())!=0);
		}
		if (p==0)
		{
			if (pp==PAWN && ((hcolor==BLACK && curY==0) || (hcolor!=BLACK && curY==7)))
			{
				pp = QUEEN;
				p = board[curX+(curY<<3)];
				while(true)
				{
					board[curX+(curY<<3)] = pp|hcolor;
					showBoard();
					key = gp_getButton();
					if (key & (BUTTON_LEFT | BUTTON_UP))
						pp = pp==ROOK ? QUEEN : pp-1;
					if (key & (BUTTON_RIGHT | BUTTON_DOWN))
						pp = pp==QUEEN ? ROOK : pp+1;
					if (key & (BUTTON_A | BUTTON_B))
						break;
					while ((key=gp_getButton())!=0);
				}
				board[curX+(curY<<3)] = p;
			}
			else
				pp = 0;
			if ((curX!=oldX || curY!=oldY) && isMoveValid(oldX,oldY,curX,curY,pp))
				break;
		}
		oldX = -1;
	}

	if (mode==1)
	{
		time_t1 = gp_getMyRTC();
		if (hcolor==BLACK)
			timeUpdate(&btime);
		else
			timeUpdate(&wtime);
	}

	lastX0 = oldX;
	lastY0 = oldY;
	lastX1 = curX;
	lastY1 = curY;
	last = NULL;
	p = strlen(game);
	game[p++] = oldX+'a';
	game[p++] = oldY+'1';
	game[p++] = curX+'a';
	game[p++] = curY+'1';
	game[p++] = ' ';
	game[p]   = 0;
	curX = oldX = -1;
}

void endGame(char* reason)
{
	int key;

	if (mode==1 && ((hcolor==BLACK && btime==0) || (hcolor!=BLACK && wtime==0)))
		reason = "TIME ELAPSED ! You lose !";

	stepProto = 3;
	last = reason;
	showBoard();

	while (true)
	{
		key=gp_getButton();
		if (key & (BUTTON_START | BUTTON_A | BUTTON_B | BUTTON_SELECT))
			break;
	}
	while ((key=gp_getButton())!=0);

	gp_mainMenu();
}

// Main loop

void gp_main()
{
	char* line;
	int i;
	switch (stepProto)
	{
		case 0: // Send uci
			gp_write("uci");
			stepProto=1;
			break;
		case 1: // Read engine info
			line = gp_read();
			if (line!=NULL && strcmp(line,"uciok")==0)
			{
				gp_write("isready");
				stepProto=2;
			}
			break;
		case 2: // Wait engine ok
			line = gp_read();
			if (line!=NULL && strcmp(line,"readyok")==0)
				stepProto=3;
			break;
		case 3: // Init game
			initBoard();
			gp_write("ucinewgame");
			if (hcolor==BLACK)
				computerMove("position startpos");
			else
				stepProto=4;
			break;
		case 4: // human move
			humanMove();
			if (stepProto!=6) // not aborted
				computerMove(game);
			break;
		case 5: // wait computer move
			line = gp_read();
			if (line!=NULL && strncmp(line,"bestmove",8)==0)
			{
				if (mode==1)
				{
					time_t1 = gp_getMyRTC();
					if (hcolor==BLACK)
						timeUpdate(&wtime);
					else
						timeUpdate(&btime);
				}
//				last = line;
				line+=9;
				if (strncmp(line,"a1a1",4)==0)
				{
					endGame("CHECKMATE ! You won !");
				}
				else
				{
					playMove(lastX0=(line[0]-'a'),lastY0=(line[1]-'1'),
							 lastX1=(line[2]-'a'),lastY1=(line[3]-'1'),line[4]);
					for (i=4;line[i] && line[i]!=' ';i++);
					if (!line[i])
					{
						line[i] = ' ';
						line[i+1] = ' ';
					}
					strncat(game,line,i+1);
					if (line[i+1]=='#')
						endGame("CHECKMATE ! You lose !");
					else if (line[i+1]=='=')
						endGame("DRAW GAME");
					else if (mode==1 && ((hcolor==BLACK && wtime==0) || (hcolor!=BLACK && btime==0)))
						endGame("TIME ELAPSED ! You win !");
					else
						stepProto = 4;
				}
			}
			break;
		case 6: // Game aborted
			while (gp_read()!=NULL);
			endGame("GAME ABORTED");
			break;
	}
}

void gp_initAll()
{
	gp_initButton();
	EnableCache();
	initGfx();
	cpu = 66;
	stepProto = 0;
	gp_initMyRTC();
	gp_mainMenu();
}
